from argparse import ArgumentParser
from os import getuid
from rbtenums import *
from rbtutils import *
from rbtexceptions import RbtConfigError

def __create_parser(param_required=False, need_create_additional_status_param=False):
    """
    Создать парсер для скрпита
    :param param_required: флаг, 
    :return:
    """
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest='commands')
    configure_parser = subparsers.add_parser(OPERATION_CONFIGURE)
    configure_parser.add_argument('--mode', action='store', required=True, help='Указать режим конфигурирования')
    if param_required:
        configure_parser.add_argument('--param', action='store', required=param_required,
                                      help='Указать аргумент скрипту, специфичный для данного режима '
                                           'в формате key1=value1/key2=value2/...')
    configure_parser.add_argument('--auth', action='store', required=False, default=str(),
                                  help='Указать параметры аутентификации (для режима ald) '
                                       'в формате username:password')

    reset_parser = subparsers.add_parser(OPERATION_RESET)
    reset_parser.add_argument('--mode', action='store', required=True, help='Указать режим деконфигурирования')
    if param_required:
        reset_parser.add_argument('--param', action='store', required=param_required,
                                  help='Указать аргумент скрипту, специфичный для данного режима '
                                       'в формате key1=value1/key2=value2/...')
    reset_parser.add_argument('--auth', action='store', required=False, default=str(),
                              help='Указать параметры аутентификации (для режима ald) '
                                   'в формате username:password')

    modes_parser = subparsers.add_parser(OPERATION_MODES)
    modes_parser.add_argument('--show', action='store_true', required=True, help='Показать доступные режимы настройки')

    status_parser = subparsers.add_parser(OPERATION_STATUS)
    status_parser.add_argument('--more', action='store_true', required=False, default=False,
                               help='Показать названия операций')
    status_parser.add_argument('--show', action='store_true', required=True,
                               help='Показать статус настройки сервиса в поддерживаемых режимах')
    status_parser.add_argument('--auth', action='store', required=False, default=str(),
                               help='Указать параметры аутентификации (для режима ald) в формате username:password.')
    if param_required:
        status_parser.add_argument('--param', action='store', required=param_required,
                                   help='Указать аргумент скрипту, специфичный для данного режима '
                                        'в формате key1=value1/key2=value2/...')
    if need_create_additional_status_param:
        modes_parser.add_argument('--service_name', action='store', required=True,
                                  help='Указать имя сервиса для просмотра его состояния')
        status_parser.add_argument('--service_name', action='store', required=True,
                                   help='Указать имя сервиса для просмотра его состояния')
        services_list_parser = subparsers.add_parser(OPERATION_SERVICES_LIST)
        services_list_parser.add_argument('--show', action='store_true', required=True,
                                          help='Просмотреть список сервисов')
    return parser


def __parse_param(param):
    """
    Парсит аргумент, конвертируя строку в словарь ключ-значение
    :param param: Параметр вида key1=value1/key2=value2 (str)
    :return: dict
    """
    def parse_key_value(row):
        key_values = row.split('=')
        if len(key_values) == 1:
            exit_with_error('Ошибка разбора аргумента скрипта')
        else:
            return {key_values[0]: key_values[1]}

    if param is None:
        return None
    result = {}
    params = param.split('/')
    if len(params) > 1:
        for p in params:
            result.update(parse_key_value(p))
    else:
        return parse_key_value(param)
    return result


def __check_auth_param(arguments):
    """
    Проверяет наличие параметра 'mode' для режима ALD
    :param arguments:
    """
    if hasattr(arguments, 'mode') and arguments.mode == MODE_ALD:
        if not arguments.auth:
            exit_with_error('Требуется строка подключения к серверу ALD')


def __get_auth_param(arguments):
    """
    Получает параметры аутентификации (если они имеются)
    :param arguments: ArgumentParser data
    :return: dict or None
    """
    if hasattr(arguments, 'auth'):
        return arguments.auth if arguments.auth is not str() else None
    return None


def __check_privileges():
    """
    Проверяет привилегии для запуска сервиса
    """
    if not getuid() == 0:
        exit_with_error('Необходимы полномочия суперпользователя для запуска')


def __pre_run(services, param_required=False):
    """
    Функция выполняет подготовительные действия для запуска задания
    :param services: list of class or class
    :param param_required: требуется параметр для param для передачи дополнительных данных скрипту (bool)
    :return: tuple(parser, args: dict)
    """
    __check_privileges()
    need_create_additional_status_param = False
    if isinstance(services, list):
        need_create_additional_status_param = True
    parser = __create_parser(param_required=param_required,
                             need_create_additional_status_param=need_create_additional_status_param)
    arguments = parser.parse_args()
    __check_auth_param(arguments)
    args = {AUTH_DATA: __get_auth_param(arguments)}
    # Разбираем параметры
    if param_required:
        args.update(__parse_param(arguments.param))
    return parser, arguments, args


def run(service_class, param_required=False):
    """
    Базовый метод запуска обработчика скрипта
    :param service_class: Класс(ы) скрипта (class)
    :param param_required: True, если скрипту необходимы дополнительные параметры (bool)
    :return:
    """
    parser, arguments, args = __pre_run(service_class, param_required)
    if isinstance(service_class, list):
        return __run_more(service_class, parser, arguments, args)
    try:
        # Создаем класс сервиса
        service = service_class(args)
        # Обрабатываем аргументы
        if arguments.commands == OPERATION_CONFIGURE:
            if arguments.mode == MODE_PAM:
                service.configure_service_pam()
            elif arguments.mode == MODE_ALD:
                service.configure_service_ald()
        elif arguments.commands == OPERATION_RESET:
            if arguments.mode == MODE_PAM:
                service.reset_service_pam()
            elif arguments.mode == MODE_ALD:
                service.reset_service_ald()
        elif arguments.commands == OPERATION_MODES:
            print(service.available_modes())
        elif arguments.commands == OPERATION_STATUS:
            print(service.status(arguments.more))
        else:
            parser.print_help()
    except RbtConfigError as error:
        exit_with_error(error.message())


def __run_more(service_classes, parser, arguments, args):
    """
    Метод для запуска обработчика скриптов для нескольких классов.
    :param service_classes: list of class
    :param arguments: Результат работы parser.parse_args()
    :param args: Словарь параметров, передаваемых скрипту (dict)
    :param parser: ArgumentParser
    :return:
    """
    class ExecutionStatus:
        def __init__(self, service_class, inverse_func):
            """
            :param service_class: class
            :param inverse_func: function
            :return:
            """
            self.service_class = service_class
            self.inverse_func = inverse_func

        def __repr__(self):
            return str(self.service_class)

    def service_names(services):
        """
        Получить имена сервисов
        :param services: list of classes
        :return: list
        """
        return [service.service_name for service in services]

    services = []
    # Создаем классы
    try:
        for service_class in service_classes:
            _service = service_class(args)
            services.append(_service)
    except RbtConfigError as error:
        exit_with_error(error.message())
    if arguments.commands == OPERATION_SERVICES_LIST:
        result = '\n'.join(service_names(services))
        result = result.rstrip('\n')
        print_stdout(result)
    elif arguments.commands == OPERATION_MODES:
        service_names = service_names(services)
        if arguments.service_name in service_names:
            service_index = service_names.index(arguments.service_name)
            result = 'Доступные режимы сервиса %s\n%s' % (arguments.service_name,
                                                          services[service_index].available_modes())
            print_stdout(result)
        else:
            exit_with_error('Отсутствует сервис \'%s\' в списке сервисов' % arguments.service_name)
    elif arguments.commands == OPERATION_STATUS:
        service_names = service_names(services)
        if arguments.service_name in service_names:
            service_index = service_names.index(arguments.service_name)
            result = 'Доступные режимы сервиса %s\n%s' % (arguments.service_name,
                                                          services[service_index].status(arguments.more))
            print_stdout(result)
        else:
            exit_with_error('Отсутствует сервис \'%s\' в списке сервисов' % arguments.service_name)
    elif arguments.commands not in [OPERATION_CONFIGURE, OPERATION_RESET]:
        parser.print_help()
        exit(1)
    # Выполняем настройку
    try:
        stack = []
        #Проверка доступных пакетов для выполнения действий для всех режимов
        for service in services:
            service.check_required_packages(arguments.mode, arguments.commands)
        #Выполнение необходимых действий всех сервисов
        for service in services:
            if arguments.commands == OPERATION_CONFIGURE:
                if arguments.mode == MODE_PAM:
                    if not service.check_pam() == STATUS_OK:
                        stack.append(ExecutionStatus(service, service.reset_service_pam))
                        service.configure_service_pam()
                elif arguments.mode == MODE_ALD:
                    if not service.check_ald() == STATUS_OK:
                        stack.append(ExecutionStatus(service, service.reset_service_ald))
                        service.configure_service_ald()
            elif arguments.commands == OPERATION_RESET:
                if arguments.mode == MODE_PAM:
                    if service.check_pam() == STATUS_OK:
                        stack.append(ExecutionStatus(service, service.configure_service_pam))
                        service.reset_service_pam()
                elif arguments.mode == MODE_ALD:
                    if service.check_ald() == STATUS_OK:
                        stack.append(ExecutionStatus(service, service.configure_service_ald))
                        service.reset_service_ald()
        if not stack:
            operation = str()
            mode = str()
            if arguments.commands == OPERATION_CONFIGURE:
                operation = 'сконфигурированы в режиме'
            else:
                operation = 'деконфигурированы из режима'
            if arguments.mode == MODE_PAM:
                mode = 'pam'
            elif arguments.mode == MODE_ALD:
                mode = 'ald'
            raise RbtConfigError('Сервисы imap и smtp уже %s %s' % (operation, mode))
    except RbtConfigError as error:
        errors = list()
        errors.append(error.message())
        for _service in reversed(stack):
            try:
                _service.inverse_func()
            except RbtConfigError as error:
                errors.append(error.message())
        exit_with_error('\n'.join(errors))
