from json import dumps
from types import FunctionType, MethodType
from rbtcommonconfigparser import RbtCommonConfigParser
from rbtutils import *
from rbtenums import *

class _ServiceConfigurator:
    def __init__(self, service, mode, operation_type, check_function, function):
        self.__service = service
        self.__mode = mode
        self.__operation = operation_type
        self.__check_function = check_function
        self.__function = function

    def run(self):
        run_condition = self.__check_function() #заменить
        if self.__operation == OPERATION_CONFIGURE and run_condition == STATUS_OK:
            raise RbtConfigError('Сервис %s уже сконфигурирован в режиме %s' % (self.__service, self.__mode))
        elif self.__operation == OPERATION_RESET and run_condition == STATUS_NOT_SET:
            raise RbtConfigError('Сервис %s уже деконфигурирован из режима %s' % (self.__service, self.__mode))
        else: 
            try:
                self.__function()        
            except RbtConfigError as error:
                if self.__operation == OPERATION_CONFIGURE:
                    error.add('Ошибка конфигурирования сервиса %s' % self.__service)
                elif self.__operation == OPERATION_RESET:
                    error.add('Ошибка деконфигурирования сервиса %s' % self.__service)
                raise RbtConfigError(error.message())            

class _DirectInverseFunctionsCall:
    """
    Класс для хранения прямого и обратного указателей на функции, а также их аргументов.
    Предполагается, что прямой и обратный указатель на функции имеют одинаковый список аргументов.
    """
    def __init__(self, direct, inverse, direct_args=None, inverse_args=None):
        """
        :param direct: function
        :param inverse: function
        :param args: tuple
        """
        self.__check_args(direct, inverse, direct_args, inverse_args)
        self.direct = direct
        self.inverse = inverse
        if direct_args:
            self.direct_args = direct_args
        else:
            self.direct_args = []
        if inverse_args:
            self.inverse_args = inverse_args
        else:
            self.inverse_args = None

    @staticmethod
    def __check_args(direct, inverse, direct_args, inverse_args):
        message = str()
        if not (isinstance(direct, MethodType) or isinstance(direct, FunctionType)):
            message += '\'direct\' имеет тип %s, требуется %s\n' % (type(direct), FunctionType)
        if not (isinstance(inverse, MethodType) or isinstance(inverse, FunctionType)):
            message += '\'inverse\' имеет тип %s, требуется %s\n' % (type(direct), FunctionType)
        if not (isinstance(direct_args, list) or direct_args is None):
            message += '\'args\' имеет тип %s, требуется %s или %s\n' % (type(direct_args), list, None)
        if not (isinstance(inverse_args, list) or inverse_args is None):
            message += '\'inverse_args\' имеет тип %s, требуется %s или %s\n' % (type(inverse_args), list, None)
        if message:
            message.rstrip('\n')
            raise RbtConfigError(message)


class BaseScript:
    __DISPLAY_MODES = {
        MODE_PAM: 'PAM',
        MODE_ALD: 'ЕПП'
    }

    def __init__(self,
                 service_name,
                 service_owner,
                 service_config_dir,
                 available_modes=None,
                 params=None,
                 ):
        """
        Конструктор класса BaseScript
        :param service_name: имя сервиса (str)
        :param service_owner: владелец сервиса (str)
        :param service_config_dir: конфигурационная директория сервиса (str)
        :param available_modes: Список операций, доступные сервису (list)
        """
        self.service_name = service_name
        self.service_owner = service_owner
        self.service_config_dir = service_config_dir
        self.__available_modes = available_modes
        self.__checks = {}
        self.__required_packages = {
            MODE_PAM: [],
            MODE_ALD: ['ald-admin', 'ald-client'],
        }
        # Переменная для стека функций, которые были исполнены при конфигурировании/деконфигурировании
        self.__functions_calls = []
        # Парсинг опций
        self.__parse_params(params)
        # Создаем локальные, ALD утилиты, а также парсер конфигурационных параметров
        self.local_utils = LocalUtils(self.service_owner)
        chown_keytab_owner = True
        keytab_directory = self.service_config_dir
        keytab_file_name = None
        auth_data = None
        principal_name = self.service_name
        if hasattr(self, AUTH_DATA):
            auth_data = self.auth_data
            del self.auth_data
        if hasattr(self, KEYTAB_DIRECTORY):
            keytab_directory = self.keytab_directory
        if hasattr(self, KEYTAB_FILE_NAME):
            keytab_file_name = self.keytab_file_name
        if hasattr(self, KRB_PRINCIPAL_NAME):
            principal_name = self.krb_principal_name
        if hasattr(self, CHOWN_KEYTAB_OWNER):
            chown_keytab_owner = self.chown_keytab_owner
        if not hasattr(self, INIT_D_SERVICE_NAME):
            self.init_d_service_name = service_name
        self.ald_utils = AldUtils(service_owner=service_owner,
                                  keytab_directory=keytab_directory,
                                  principal_name=principal_name,
                                  keytab_file_name=keytab_file_name,
                                  auth_data=auth_data,
                                  chown_keytab_owner=chown_keytab_owner)
        self.config_parser = RbtCommonConfigParser(service_config_dir)

        # Получение статуса настроек сервиса
        self.__get_checks()

    def __get_checks(self):
        """
        Служебная функция по заполенению словаря checks
        """
        if self.__available_modes is not None:
            if MODE_PAM in self.__available_modes:
                self.__checks[MODE_PAM] = self.check_pam()
            if MODE_ALD in self.__available_modes:
                self.__checks[MODE_ALD] = self.check_ald()

    def __get_check(self, key):
        """
        Получает режим настройки сервиса в режиме :param key
        :param key: str
        :return:str
        """
        try:
            return self.__checks[key]
        except KeyError:
            return STATUS_NOT_SET

    def __parse_params(self, params):
        if params:
            for key in params:
                if not key == REQUIRED_PACKAGES:
                    setattr(self, '%s' % key, params[key])
                else:
                    self.__set_required_packages(params[key])

    def __set_required_packages(self, packages):
        for mode in packages:
            for package in packages[mode]:
                if package not in self.__required_packages[mode]:
                    self.__required_packages[mode] += [package]
    
    def __check_required_packages(self, mode, operation):
        """
        Проверяет наличие требуемых пакетов для конфигурирования/деконфигурирования сервиса
        :param mode: Режим работы (str)
        :param operation: Тип операции (str)
        :raise RbtConfigError, если не все требуемые пакеты установлены
        :return: None
        """
        uninstalled_packages = []
        package_list = self.__required_packages[mode]
        for package in package_list:
            if not self.local_utils.check_if_package_exists(package):
                uninstalled_packages.append(package)
        # Формируем сообщение о неустановленных пакетах
        if not len(uninstalled_packages) == 0:
            packages = str()
            message = str()
            for package in uninstalled_packages:
                packages += '  %s\n' % package
            if operation == OPERATION_CONFIGURE:
                message = 'Для конфигурирования сервиса в режиме %s требуются следующие пакеты: \n%s'\
                          % (self.__DISPLAY_MODES[mode],
                             packages)
            elif operation == OPERATION_RESET:
                message = 'Для деконфигурирования сервиса из режима %s требуются следующие пакеты: \n%s'\
                          % (self.__DISPLAY_MODES[mode],
                             packages)
            raise RbtConfigError(message)

    def __print_configure_message(self, mode):
        """
        Распечатывает сообщение, если оно есть, в случае успешной конфигурации сервиса в режиме :param mode.
        :param mode: режим настройки (str)
        """
        if not hasattr(self, MESSAGES):
            return
        if self.messages is not None:
            try:
                print_stderr(self.messages[CONFIGURE][mode])
            except KeyError:
                pass

    def __print_reset_message(self, mode):
        """
        Распечатывает сообщение, если оно есть, в случае успешной деконфигурации сервиса в режиме :param mode.
        :param mode: режим настройки (str)
        """
        if not hasattr(self, MESSAGES):
            return
        if self.messages is not None:
            try:
                print_stderr(self.messages[RESET][mode])
            except KeyError:
                pass

    def __check_is_modes_incompatible(self, mode):
        """
        Проверяет, возможна ли настройка сервиса в режиме :param mode
        :param mode: str
        :return: bool
        """
        if hasattr(self, INCOMPATIBLE_MODES):
            for item in self.incompatible_modes:
                if (self.__checks[item[0]] == STATUS_OK and item[1] == mode) or \
                        (self.__checks[item[1]] == STATUS_OK and item[0] == mode):
                    return False, item
        return True, None

    def __exec_reverse_functions(self):
        """
        Раскручивает стек примененных операций, восстанавливая исходное состояние системы
        :return:
        """
        for function in reversed(self.__functions_calls):
            try:
                if function.inverse_args:
                    args = function.inverse_args
                else:
                    args = function.direct_args
                function.inverse(*args)
            except RbtConfigError:
                pass
            
    def restart_service(self, raise_exception_if_error=True):
        """
        Выполняет перезапуск сервиса
        :return: None
        """
        try:
            self.local_utils.exec_command('service %s restart' % self.init_d_service_name, enable_stderr=True)
        except RbtConfigError as error:
            if raise_exception_if_error:
                raise RbtConfigError(error.message())

    def exec_function_with_inverse_call(self, direct, inverse, direct_args=None, inverse_args=None):
        """
        Функция помещает указатели на функции :param direct и :param inverse в стек с аргументами :param args.
        Пытается выполнить её.
        :param direct: function
        :param inverse: function
        :param direct_args: tuple
        :param inverse_args: tuple
        :return: None
        """
        call = _DirectInverseFunctionsCall(direct, inverse, direct_args, inverse_args)
        try:
            call.direct(*call.direct_args)
        except RbtConfigError as error:
            raise RbtConfigError(error.message())
        else:
            # Заносим функцию, обратную функцию и аргументы в стек
            self.__functions_calls.append(call)

    def check_pam(self):
        """
        Функция проверки, настроен ли сервис в режиме работы PAM
        Должна быть переопределена в классах-наследниках.
        Функция должна выводить одно из следующих значений:
        STATUS_OK, STATUS_ERROR, STATUS_NOT_SET -- см. rbtenums.py.
        :return str
        """
        return STATUS_NOT_SET

    def configure_pam(self):
        """
        Содержит операции по настройке сервиса в режиме PAM
        Должа быть переопределена в классах-наследниках
        """
        pass

    def reset_pam(self):
        """
        Содержит операции по деконфигурирования сервиса в режиме PAM
        Должна быть переопределена в классах-наследниках
        """
        pass

    def check_required_packages(self, mode, operation):
        self.__check_required_packages(mode, operation)

    def configure_service_pam(self):
        """
        Функция настройки сервиса в режиме PAM
        Для настройки используется интерфейсная функция configure_pam, которая выполняет настройку сервиса
        """
        # Проверяем, настроен ли сервис в каком-либо режиме. И совместимы ли они?
        check_status = self.__check_is_modes_incompatible(MODE_PAM)
        if check_status[0]:
            # Проверяем наличие необходимых пакетов
            self.__check_required_packages(MODE_PAM, OPERATION_CONFIGURE)
            # Создаем объект, с помощью которого реализуется запуск задания
            configurator = _ServiceConfigurator(service=self.service_name,
                                                mode=self.__DISPLAY_MODES[MODE_PAM],
                                                operation_type=OPERATION_CONFIGURE,
                                                check_function=self.check_pam,
                                                function=self.configure_pam)
            try:
                configurator.run()
                # Перезапускаем сервис
                self.restart_service()
                #Выводим модель
                self.__print_configure_message(MODE_PAM)
            except RbtConfigError as error:
                self.__exec_reverse_functions()
                # Восстанавливаем бекап
                if not self.__checks[MODE_PAM] == 'OK':
                    self.config_parser.restore_backup()
                # Перезапускаем сервис, не выбрасывая возможного исключения
                self.restart_service(raise_exception_if_error=False)
                raise RbtConfigError(error.message())
            finally:
                # Очищаем стек вызова функций
                self.__functions_calls = []
                #Удаляем бекап
                self.config_parser.del_backup()                
        else:
            incompatible_modes = check_status[1]
            current_working_mode = MODE_ALD if incompatible_modes[0] == MODE_ALD else incompatible_modes[1]
            raise RbtConfigError('Сервис одновременно не может работать в режимах %s и %s.\n'
                                 'Выполните деконфигурацию из режима %s, после чего повторите конфигурацию сервиса '
                                 'в режиме %s' % (self.__DISPLAY_MODES[incompatible_modes[0]],
                                                  self.__DISPLAY_MODES[incompatible_modes[1]],
                                                  self.__DISPLAY_MODES[current_working_mode],
                                                  self.__DISPLAY_MODES[MODE_ALD]))

    def reset_service_pam(self):
        """
        Функция деконфигурирования сервиса в режиме PAM
        Для деконфигурирования используется интерфейсная функция reset_pam, которая выполняет деконфигурирование сервиса
        """
        # Проверяем наличие необходимых пакетов
        self.__check_required_packages(MODE_PAM, OPERATION_RESET)
        # Создаем объект, с помощью которого реализуется запуск задания
        configurator = _ServiceConfigurator(service=self.service_name,
                                            mode=self.__DISPLAY_MODES[MODE_PAM],
                                            operation_type=OPERATION_RESET,
                                            check_function=self.check_pam,
                                            function=self.reset_pam
                                            )
        try:
            configurator.run()
            # Перегружаем сервис
            self.restart_service()
            self.__print_reset_message(MODE_PAM)
        except RbtConfigError as error:         
            self.__exec_reverse_functions()
            # Перезапускаем сервис, не выбрасывая возможного исключения
            self.restart_service(raise_exception_if_error=False)
            raise RbtConfigError(error.message())
        finally:
            # Очищаем стек вызова функций
            self.__functions_calls = []
            #Удаляем бекап
            self.config_parser.del_backup()            

    def check_ald(self):
        """
        Функция проверки, настроен ли сервис в режиме работы ALD
        Должна быть переопределена в классах-наследниках
        Функция должна выводить одно из следующих значений:
        STATUS_OK, STATUS_ERROR, STATUS_NOT_SET -- см. rbtenums.py
        :return str
        """
        return STATUS_NOT_SET

    def configure_ald(self):
        """
        Содержит операции по настройке сервиса в режиме ЕПП
        Должа быть переопределена в классах-наследниках
        """
        pass

    def reset_ald(self):
        """
        Содержит операции по деконфигурирования сервиса в режиме PAM
        Должна быть переопределена в классах-наследниках
        """
        pass

    def configure_service_ald(self):
        """
        Функция настройки сервиса в режиме ЕПП
        Для настройки используется интерфейсная функция configure_ald, которая выполняет настройку сервиса
        """
        check_status = self.__check_is_modes_incompatible(MODE_ALD)
        if check_status[0]:
            # Выполняем проверку, запущен ли ALD
            if not self.ald_utils.check_is_ald_running():
                raise RbtConfigError('ALD не настроен или выключен')
            # Проверяем наличие необходимых пакетов
            self.__check_required_packages(MODE_ALD, OPERATION_CONFIGURE)
            # Создаем объект, с помощью которого реализуется запуск задания
            configurator = _ServiceConfigurator(service=self.service_name,
                                                mode=self.__DISPLAY_MODES[MODE_ALD],
                                                operation_type=OPERATION_CONFIGURE,
                                                check_function=self.check_ald,
                                                function=self.configure_ald
                                                )
            try:
                configurator.run()
                # Перегружаем сервис
                self.restart_service()
                self.__print_configure_message(MODE_ALD)
            except RbtConfigError as error:
                self.__exec_reverse_functions()
                # Восстанавливаем бекап
                if not self.__checks[MODE_ALD] == 'OK':
                    self.config_parser.restore_backup()
                # Перезапускаем сервис, не выбрасывая возможного исключения
                self.restart_service(raise_exception_if_error=False)
                raise RbtConfigError(error.message())
            finally:
                # Очищаем стек вызова функций
                self.__functions_calls = []
                #Удаляем бекап
                self.config_parser.del_backup()
        else:
            incompatible_modes = check_status[1]
            current_working_mode = MODE_PAM if incompatible_modes[0] == MODE_PAM else incompatible_modes[1]
            raise RbtConfigError('Сервис одновременно не может работать в режимах %s и %s.\n'
                                 'Выполните деконфигурацию из режима %s, после чего повторите конфигурацию сервиса '
                                 'в режиме %s' % (self.__DISPLAY_MODES[incompatible_modes[0]],
                                                  self.__DISPLAY_MODES[incompatible_modes[1]],
                                                  self.__DISPLAY_MODES[current_working_mode],
                                                  self.__DISPLAY_MODES[MODE_ALD]))

    def reset_service_ald(self):
        """
        Функция деконфигурирования сервиса в режиме ЕПП
        Для деконфигурирования используется интерфейсная функция reset_ald, которая выполняет деконфигурирование сервиса
        """
        # Выполняем проверку, запущен ли ALD
        if not self.ald_utils.check_is_ald_running():
            raise RbtConfigError('ALD не настроен или выключен')
        self.__check_required_packages(MODE_ALD, OPERATION_RESET)
        configurator = _ServiceConfigurator(service=self.service_name,
                                            mode=self.__DISPLAY_MODES[MODE_ALD],
                                            operation_type=OPERATION_RESET,
                                            check_function=self.check_ald,
                                            function=self.reset_ald)
        try:
            configurator.run()
            # Перезапускаем сервис
            self.restart_service()
        except RbtConfigError as error:
            self.__exec_reverse_functions()
            # Перезапускам сервис, не выбрасывая возможного исключения
            self.restart_service(raise_exception_if_error=False)
            raise RbtConfigError(error.message())
        finally:
            # Очищаем стек вызова функций
            self.__functions_calls = []
            #Удаляем бекап
            self.config_parser.del_backup()            

    def available_modes(self):
        """
        Возваращет список операций, в которых сервис может быть сконфигурирован/деконфигурирован.
        :return: str
        """
        if self.available_modes is not None:
            result = str()
            for mode in self.__available_modes:
                result += '%s\n' % mode
            return result.rstrip('\n')
        return None

    def status(self, more=False):
        """
        :param more: Если True, то выдается в расширенном формате
        Возвращает статус сервиса в следующем виде в формате JSON
        """
        # Получаем стандартные расширенные описания режимов работы
        descriptions = dict((key, MODES[key][1]) for key in MODES)
        # Есть другие? Обновляем ключи
        if hasattr(self, EXTENDED_DESCRIPTIONS):
            for key in self.extended_descriptions:
                descriptions[key] = self.extended_descriptions[key]
        result = {}
        if hasattr(self, INCOMPATIBLE_MODES):
            result[INCOMPATIBLE_MODES] = self.incompatible_modes
        else:
            result[INCOMPATIBLE_MODES] = []
        result['modes'] = []
        for key in self.__checks:
            if more:
                mode_info = {
                    'header': MODES[key][0],
                    'description': descriptions[key],
                    'status': self.__checks[key]
                }
            else:
                mode_info = {
                    'status': self.__checks[key]
                }
            result['modes'] += [{key: mode_info}]
        return dumps(result, ensure_ascii=False, indent=4)
