#!/usr/bin/env python3
from __init__ import *

class Apache(BaseScript):
    MODE_PAM_REQUIRED_PACKAGE = 'libapache2-mod-authnz-pam'
    MODE_ALD_REQUIRED_PACKAGE = 'libapache2-mod-auth-kerb'

    def __init__(self, params):
        service_name = 'apache'
        service_config_dir = '/etc/apache2'
        service_owner = 'www-data'
        available_modes = [MODE_PAM, MODE_ALD]
        params.update({
            EXTENDED_DESCRIPTIONS: {
                MODE_PAM: '%s для сайта \'default\'' % MODE_PAM_EXTENDED_DESCRIPTION,
                MODE_ALD: '%s для сайта \'default\' с параметром safe_credential = on' % MODE_ALD_EXTENDED_DESCRIPTION
            },
            REQUIRED_PACKAGES: {
                MODE_PAM: [self.MODE_PAM_REQUIRED_PACKAGE],
                MODE_ALD: [self.MODE_ALD_REQUIRED_PACKAGE]
            },
            INCOMPATIBLE_MODES: [(MODE_ALD, MODE_PAM)],
            INIT_D_SERVICE_NAME: 'apache2',
            KEYTAB_FILE_NAME: 'keytab',
            KRB_PRINCIPAL_NAME: 'HTTP'
        })
        self.__default_host_file = '/etc/apache2/sites-available/000-default.conf'
        self.__pam_script = '/etc/pam.d/apache2'
        self.__faillog_path = '/var/log/faillog'
        self.__pam_script_data = ['account required pam_tally.so']
        self.__virtual_host_data = {
            MODE_PAM: [
                'AuthType Basic', 
                'AuthName "Pam Auth"',
                'AuthBasicProvider PAM', 
                'AuthPAMService apache2',
                'require valid-user'
            ],
            MODE_ALD: [
                'AuthType Kerberos',
                'KrbAuthRealms %s' % AldUtils.realm(),
                'KrbServiceName HTTP/%s' % BaseUtils.hostname(),
                'Krb5Keytab %s/keytab' % service_config_dir,
                'KrbMethodNegotiate on',
                'KrbMethodK5Passwd off',
                'KrbSaveCredentials on',
                'require valid-user'
            ]
        }
        super().__init__(service_name,
                         service_owner,
                         service_config_dir=service_config_dir,
                         available_modes=available_modes,
                         params=params
                         )

    def a2enmod(self, module):
        return self.local_utils.exec_command('a2enmod %s' % module, enable_stderr=True)

    def a2dismod(self, module):
        return self.local_utils.exec_command('a2dismod %s' % module, enable_stderr=True)

    def check_is_module_loaded(self, module):
        """
        Проверяет, запущен ли модуль :param module
        :param module: str
        :return: bool
        """
        try:
            self.local_utils.exec_command('apache2ctl -t -D DUMP_MODULES | grep %s' % module)
        except RbtConfigError:
            return False
        return True
    
    def run_service(self, service):
        try:
            self.local_utils.exec_command('service %s start' % service)
        except RbtConfigError:
            return False
        return True    

    def check_pam(self):
        checks = list()
        checks.append(self.check_is_module_loaded('authnz_pam'))
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   ['<VirtualHost *:80>']))
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=['</VirtualHost>'],
                                                                   after='<VirtualHost *:80>'))        
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=['ServerName %s' % self.local_utils.hostname()],
                                                                   after='<VirtualHost *:80>',
                                                                   before='</VirtualHost'))
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=['<Directory /var/www/>',
                                                                         '</Directory>'],
                                                                   after='<VirtualHost *:80>',
                                                                   before='</VirtualHost'))
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=['</Directory>'],
                                                                   after='<Directory /var/www/>'))        
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=self.__virtual_host_data[MODE_PAM],
                                                                   after='<Directory /var/www/>',
                                                                   before='</Directory>'))
        checks.append(self.local_utils.check_facl_on_parsec_db())
        checks.append(self.local_utils.check_is_user_in_shadow_group())
        checks.append(self.config_parser.config_file_contains_rows(self.__pam_script, self.__pam_script_data))
        checks.append(self.local_utils.check_facl(self.__faillog_path, 'rw'))
        return analyze_check(checks)

    def configure_pam(self):
        if not self.check_is_module_loaded('authnz_pam'):
            self.exec_function_with_inverse_call(direct=self.a2enmod,
                                                 inverse=self.a2dismod,
                                                 direct_args=['authnz_pam'])
        if self.check_is_module_loaded('auth_kerb'):
            self.exec_function_with_inverse_call(direct=self.a2dismod,
                                                 inverse=self.a2enmod,
                                                 direct_args=['auth_kerb'])          
        self.config_parser.add_rows_to_config_file(self.__default_host_file,
                                                    ['<VirtualHost *:80>'])  
        self.config_parser.add_rows_after(self.__default_host_file,
                                         ['</VirtualHost>'],
                                         '<VirtualHost *:80>')
        self.config_parser.add_rows_after(self.__default_host_file,
                                         ['ServerName %s' % self.local_utils.hostname()],
                                         '<VirtualHost *:80>')
        self.config_parser.add_rows_after(self.__default_host_file,
                                         ['<Directory /var/www/>'],
                                         '<VirtualHost *:80>')
        self.config_parser.add_rows_after(self.__default_host_file,
                                         ['</Directory>'],
                                         '<Directory /var/www/>')     
        self.config_parser.add_rows_after(self.__default_host_file,
                                         self.__virtual_host_data[MODE_PAM],
                                         '<Directory /var/www/>')         
        self.exec_function_with_inverse_call(direct=self.local_utils.set_facl_on_parsec_db_to_user,
                                             inverse=self.local_utils.del_facl_on_parsec_db_from_user)
        self.exec_function_with_inverse_call(direct=self.local_utils.add_user_to_shadow_group,
                                             inverse=self.local_utils.del_user_from_shadow_group)
        self.config_parser.add_rows_to_config_file(self.__pam_script,
                                                    self.__pam_script_data)
        self.exec_function_with_inverse_call(direct=self.local_utils.set_facl,
                                             inverse=self.local_utils.del_facl,
                                             direct_args=['rw', self.__faillog_path],
                                             inverse_args=[self.__faillog_path])

    def reset_pam(self):        
        self.exec_function_with_inverse_call(direct=self.local_utils.del_facl_on_parsec_db_from_user,
                                             inverse=self.local_utils.set_facl_on_parsec_db_to_user)
        if self.local_utils.check_is_user_in_shadow_group():
            self.exec_function_with_inverse_call(direct=self.local_utils.del_user_from_shadow_group,
                                                 inverse=self.local_utils.add_user_to_shadow_group)
        self.exec_function_with_inverse_call(direct=self.a2enmod,
                                             inverse=self.a2dismod,
                                             direct_args=['authnz_pam'])
        self.config_parser.remove_rows_from_config(self.__pam_script, self.__pam_script_data)
        self.exec_function_with_inverse_call(direct=self.local_utils.del_facl,
                                             inverse=self.local_utils.set_facl,
                                             direct_args=[self.__faillog_path],
                                             inverse_args=['rw', self.__faillog_path])
        if self.local_utils.check_if_package_exists(self.MODE_ALD_REQUIRED_PACKAGE):
            self.exec_function_with_inverse_call(direct=self.a2enmod,
                                                 inverse=self.a2dismod,
                                                 direct_args=['auth_kerb'])
        self.config_parser.remove_rows_from_config(self.__default_host_file,
                                                   ['ServerName %s' % self.local_utils.hostname(),
                                                    '<VirtualHost *:80>',
                                                    '</VirtualHost>',
                                                    '<Directory /var/www/>',
                                                    '</Directory>'
                                                    ] + 
                                                   self.__virtual_host_data[MODE_PAM])        

    def check_ald(self):
        checks = list()
        checks.append(not self.check_is_module_loaded('authnz_pam'))
        checks.append(self.check_is_module_loaded('auth_kerb'))    
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   ['<VirtualHost *:80>']))
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=['</VirtualHost>'],
                                                                   after='<VirtualHost *:80>'))        
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=['ServerName %s' % self.local_utils.hostname()],
                                                                   after='<VirtualHost *:80>',
                                                                   before='</VirtualHost'))
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=['<Directory /var/www/>',
                                                                         '</Directory>'],
                                                                   after='<VirtualHost *:80>',
                                                                   before='</VirtualHost'))
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=['</Directory>'],
                                                                   after='<Directory /var/www/>'))
        checks.append(self.config_parser.config_file_contains_rows(self.__default_host_file,
                                                                   rows=self.__virtual_host_data[MODE_ALD],
                                                                   after='<Directory /var/www/>',
                                                                   before='</Directory>'))
        if self.ald_utils.check_if_ald_pass_file_exits():     
            checks.append(self.ald_utils.check_is_service_in_domain())
            checks.append(self.ald_utils.check_is_service_in_group('mac'))
        return analyze_check(checks)        

    def configure_ald(self):
        if self.check_is_module_loaded('authnz_pam'):
            self.exec_function_with_inverse_call(direct=self.a2dismod,
                                                 inverse=self.a2enmod,
                                                 direct_args=['authnz_pam'])
        if not self.check_is_module_loaded('auth_kerb'):
            self.exec_function_with_inverse_call(direct=self.a2enmod,
                                                 inverse=self.a2dismod,
                                                 direct_args=['auth_kerb'])        
        self.config_parser.add_rows_to_config_file(self.__default_host_file,
                                                    ['<VirtualHost *:80>'])  
        self.config_parser.add_rows_after(self.__default_host_file,
                                         ['</VirtualHost>'],
                                         '<VirtualHost *:80>')
        self.config_parser.add_rows_after(self.__default_host_file,
                                         ['ServerName %s' % self.local_utils.hostname()],
                                         '<VirtualHost *:80>')
        self.config_parser.add_rows_after(self.__default_host_file,
                                         ['<Directory /var/www/>'],
                                         '<VirtualHost *:80>')
        self.config_parser.add_rows_after(self.__default_host_file,
                                         ['</Directory>'],
                                         '<Directory /var/www/>')     
        self.config_parser.add_rows_after(self.__default_host_file,
                                         self.__virtual_host_data[MODE_ALD],
                                         '<Directory /var/www/>')   
        self.run_service('krb5-admin-server') #TODO: потом поправить        
        if not self.ald_utils.check_is_service_in_domain():
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_domain,
                                                 inverse=self.ald_utils.del_service_from_domain)
        self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_group,
                                             inverse=self.ald_utils.del_service_from_group,
                                             direct_args=['mac'])
        self.exec_function_with_inverse_call(direct=self.ald_utils.create_keytab_for_service,
                                             inverse=self.ald_utils.remove_keytab)

    def reset_ald(self):
        if self.local_utils.check_if_package_exists(self.MODE_PAM_REQUIRED_PACKAGE):
            self.exec_function_with_inverse_call(direct=self.a2enmod,
                                                 inverse=self.a2dismod,
                                                 direct_args=['authnz_pam'])        
            self.exec_function_with_inverse_call(direct=self.a2dismod,
                                                 inverse=self.a2enmod,
                                                 direct_args=['auth_kerb'])          
        self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_group,
                                             inverse=self.ald_utils.add_service_to_group,
                                             direct_args=['mac'])        
        self.exec_function_with_inverse_call(direct=self.ald_utils.remove_keytab,
                                             inverse=self.ald_utils.create_keytab_for_service)
        self.run_service('krb5-admin-server') #TODO: потом поправить        
        if self.ald_utils.check_is_service_in_domain():
            self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_domain,
                                                 inverse=self.ald_utils.add_service_to_domain)             
        self.config_parser.remove_rows_from_config(self.__default_host_file,
                                                   ['ServerName %s' % self.local_utils.hostname(),
                                                    '<VirtualHost *:80>',
                                                    '</VirtualHost>',
                                                    '<Directory /var/www/>',
                                                    '</Directory>'
                                                    ] + 
                                                   self.__virtual_host_data[MODE_ALD])        
if __name__ == '__main__':
    run(Apache, param_required=False)
